@@ -27,9 +27,11 @@ from nbxmpp.protocol import BindRequest
 from nbxmpp.protocol import TLSRequest
 from nbxmpp.protocol import isResultNode
 from nbxmpp.protocol import JID
-from nbxmpp.protocol import Iq
 from nbxmpp.protocol import Protocol
 from nbxmpp.protocol import WebsocketCloseHeader
+from nbxmpp.errors import TimeoutStanzaError
+from nbxmpp.errors import StanzaError
+from nbxmpp.errors import CancelledError
 from nbxmpp.addresses import ServerAddresses
 from nbxmpp.addresses import NoMoreAddresses
 from nbxmpp.tcp import TCPConnection
@@ -46,7 +48,6 @@ from nbxmpp.util import get_stream_header
 from nbxmpp.util import get_stanza_id
 from nbxmpp.util import Observable
 from nbxmpp.util import validate_stream_header
-from nbxmpp.util import is_error_result
 from nbxmpp.util import LogAdapter
 
 log = logging.getLogger('nbxmpp.stream')
@@ -95,6 +96,8 @@ class Client(Observable):
         self._allowed_protocols = None
         self._allowed_mechs = None
 
+        self._sm_disabled = False
+
         self._stream_id = None
         self._stream_secure = False
         self._stream_authenticated = False
@@ -102,10 +105,10 @@ class Client(Observable):
         self._session_required = False
         self._connect_successful = False
         self._stream_close_initiated = False
-        self._ping_id = None
+        self._ping_task = None
         self._error = None, None, None
 
-        self._ignored_tls_errors = []
+        self._ignored_tls_errors = set()
         self._ignore_tls_errors = False
         self._accepted_certificates = []
         self._peer_certificate = None
@@ -113,9 +116,11 @@ class Client(Observable):
 
         self._con = None
         self._local_address = None
+        self._remote_address = None
         self._mode = Mode.CLIENT
 
         self._ping_source_id = None
+        self._tasks = []
 
         self._dispatcher = StanzaDispatcher(self)
         self._dispatcher.subscribe('before-dispatch', self._on_before_dispatch)
@@ -127,6 +132,15 @@ class Client(Observable):
 
         self._state = StreamState.DISCONNECTED
 
+    def add_task(self, task):
+        self._tasks.append(task)
+
+    def remove_task(self, task, _context):
+        try:
+            self._tasks.remove(task)
+        except Exception:
+            pass
+
     @property
     def log_context(self):
         return self._log_context
@@ -184,6 +198,8 @@ class Client(Observable):
         return self._ignored_tls_errors
 
     def set_ignored_tls_errors(self, errors):
+        if errors is None:
+            errors = set()
         self._ignored_tls_errors = errors
 
     @property
@@ -204,6 +220,10 @@ class Client(Observable):
     def peer_certificate(self):
         return self._peer_certificate, self._peer_certificate_errors
 
+    @property
+    def current_address(self):
+        return self._current_address
+
     @property
     def current_connection_type(self):
         return self._current_address.type
@@ -242,8 +262,14 @@ class Client(Observable):
     def local_address(self):
         return self._local_address
 
+    @property
+    def remote_address(self):
+        return self._remote_address
+
     @property
     def connection_types(self):
+        if self._custom_host is not None:
+            return [self._custom_host[2]]
         return list(self._allowed_con_types or [ConnectionType.DIRECT_TLS,
                                                 ConnectionType.START_TLS])
 
@@ -254,6 +280,7 @@ class Client(Observable):
     def mechs(self):
         return set(self._allowed_mechs or set(['SCRAM-SHA-256',
                                                'SCRAM-SHA-1',
+                                               'GSSAPI',
                                                # 'SCRAM-SHA-256-PLUS',
                                                # 'SCRAM-SHA-1-PLUS',
                                                'PLAIN']))
@@ -263,12 +290,21 @@ class Client(Observable):
 
     @property
     def protocols(self):
+        if self._custom_host is not None:
+            return [self._custom_host[1]]
         return list(self._allowed_protocols or [ConnectionProtocol.TCP,
                                                 ConnectionProtocol.WEBSOCKET])
 
     def set_protocols(self, protocols):
         self._allowed_protocols = protocols
 
+    def set_sm_disabled(self, value):
+        self._sm_disabled = value
+
+    @property
+    def sm_disabled(self):
+        return self._sm_disabled
+
     @property
     def client_cert(self):
         return self._client_cert, self._client_cert_pass
@@ -289,7 +325,7 @@ class Client(Observable):
         return self._jid
 
     def _set_bound_jid(self, jid):
-        self._jid = JID(jid)
+        self._jid = JID.from_string(jid)
 
     @property
     def has_error(self):
@@ -403,7 +439,8 @@ class Client(Observable):
     def _disconnect(self, immediate=True):
         self.state = StreamState.DISCONNECTING
         self._remove_ping_timer()
-        self._dispatcher.remove_ping_callback(self._ping_id)
+        self._cancel_ping_task()
+
         if not immediate:
             self._stream_close_initiated = True
             self._smacks.close_session()
@@ -419,11 +456,14 @@ class Client(Observable):
     def _on_connected(self, connection, _signal_name):
         self.set_state(StreamState.CONNECTED)
         self._local_address = connection.local_address
+        self._remote_address = connection.remote_address
 
     def _on_disconnected(self, _connection, _signal_name):
         self.state = StreamState.DISCONNECTED
+        for task in self._tasks:
+            task.cancel()
         self._remove_ping_timer()
-        self._dispatcher.remove_ping_callback(self._ping_id)
+        self._cancel_ping_task()
         self._reset_stream()
         self.notify('disconnected')
 
@@ -457,7 +497,7 @@ class Client(Observable):
         if not self._stream_close_initiated:
             self.state = StreamState.DISCONNECTING
             self._remove_ping_timer()
-            self._dispatcher.remove_ping_callback(self._ping_id)
+            self._cancel_ping_task()
             self._smacks.close_session()
             self._end_stream()
             self._con.shutdown_output()
@@ -786,21 +826,33 @@ class Client(Observable):
 
     def _ping(self):
         self._ping_source_id = None
-        iq = Iq('get', to=self.domain)
-        iq.addChild(name='ping', namespace=Namespace.PING)
-        self._ping_id = self.send_stanza(iq,
-                                         timeout=10,
-                                         callback=self._on_pong)
-        self._log.info('Ping')
+        self._ping_task = self.get_module('Ping').ping(
+            self.domain,
+            timeout=10,
+            callback=self._on_pong)
+
+    def _on_pong(self, task):
+        self._ping_task = None
+
+        try:
+            task.finish()
+        except TimeoutStanzaError:
+            self._log.info('Ping timeout')
+            self._disconnect(immediate=True)
+            return
+
+        except CancelledError:
+            return
+
+        except StanzaError:
+            pass
 
-    def _on_pong(self, _client, result):
-        if is_error_result(result):
-            if result.condition == 'timeout':
-                self._log.info('Ping timeout')
-                self._disconnect(immediate=True)
-                return
         self._log.info('Pong')
 
+    def _cancel_ping_task(self):
+        if self._ping_task is not None:
+            self._ping_task.cancel()
+
     def register_handler(self, *args, **kwargs):
         self._dispatcher.register_handler(*args, **kwargs)
 
@@ -808,6 +860,8 @@ class Client(Observable):
         self._dispatcher.unregister_handler(*args, **kwargs)
 
     def destroy(self):
+        for task in self._tasks:
+            task.cancel()
         self._remove_ping_timer()
         self._smacks = None
         self._sasl = None
