@@ -22,7 +22,6 @@ from collections import namedtuple
 from gi.repository import Soup
 from gi.repository import Gio
 
-from nbxmpp.protocol import JID
 from nbxmpp.namespaces import Namespace
 from nbxmpp.protocol import Protocol
 from nbxmpp.protocol import Node
@@ -74,28 +73,18 @@ StanzaIDData.__new__.__defaults__ = (None, None)
 PubSubEventData = namedtuple('PubSubEventData', 'node id item data deleted retracted purged')
 PubSubEventData.__new__.__defaults__ = (None, None, None, False, False, False)
 
-PubSubConfigResult = namedtuple('PubSubConfigResult', 'jid node form')
-
-PubSubPublishResult = namedtuple('PubSubPublishResult', 'jid node id')
-
 MoodData = namedtuple('MoodData', 'mood text')
 
+BlockingPush = namedtuple('BlockingPush', 'block unblock unblock_all')
+
 ActivityData = namedtuple('ActivityData', 'activity subactivity text')
 
 LocationData = namedtuple('LocationData', LOCATION_DATA)
 LocationData.__new__.__defaults__ = (None,) * len(LocationData._fields)
 
-AvatarMetaData = namedtuple('AvatarMetaData', 'bytes height width id type url')
-AvatarMetaData.__new__.__defaults__ = (None,) * len(AvatarMetaData._fields)
-
-AvatarData = namedtuple('AvatarData', 'jid sha data')
-AvatarData.__new__.__defaults__ = (None,) * len(AvatarData._fields)
-
 BookmarkData = namedtuple('BookmarkData', 'jid name nick autojoin password')
 BookmarkData.__new__.__defaults__ = (None, None, None, None)
 
-BlockingListResult = namedtuple('BlockingListResult', 'blocking_list')
-
 PGPPublicKey = namedtuple('PGPPublicKey', 'jid key date')
 
 PGPKeyMetadata = namedtuple('PGPKeyMetadata', 'jid fingerprint date')
@@ -126,14 +115,8 @@ OOBData = namedtuple('OOBData', 'url desc')
 
 CorrectionData = namedtuple('CorrectionData', 'id')
 
-DisplayMarking = namedtuple('DisplayMarking', 'label fgcolor bgcolor')
-SecurityLabel = namedtuple('SecurityLabel', 'displaymarking')
-
 RegisterData = namedtuple('RegisterData', 'instructions form fields_form oob_url bob_data')
 
-ChangePasswordResult = namedtuple('ChangePasswordResult', 'successful form')
-ChangePasswordResult.__new__.__defaults__ = (None,)
-
 HTTPUploadData = namedtuple('HTTPUploadData', 'put_uri get_uri headers')
 HTTPUploadData.__new__.__defaults__ = (None,)
 
@@ -250,7 +233,7 @@ class DiscoInfo(namedtuple('DiscoInfo', 'stanza identities features dataforms ti
             return self.muc_identity_name
 
         if self.jid is not None:
-            return self.jid.getNode()
+            return self.jid.localpart
         return None
 
     @property
@@ -413,14 +396,14 @@ class DiscoIdentity(namedtuple('DiscoIdentity', 'category type name lang')):
         return hash(str(self))
 
 
-class AdHocCommand(namedtuple('AdHocCommand', 'jid node name sessionid status data actions notes')):
+class AdHocCommand(namedtuple('AdHocCommand', 'jid node name sessionid status data actions default notes')):
 
     __slots__ = []
 
     def __new__(cls, jid, node, name, sessionid=None, status=None,
-                data=None, actions=None, notes=None):
+                data=None, actions=None, default=None, notes=None):
         return super(AdHocCommand, cls).__new__(cls, jid, node, name, sessionid,
-                                                status, data, actions, notes)
+                                                status, data, actions, default, notes)
 
     @property
     def is_completed(self):
@@ -573,22 +556,6 @@ class StanzaMalformedError(CommonError):
         raise NotImplementedError
 
 
-class StanzaTimeoutError(CommonError):
-    def __init__(self, id_):
-        self.condition = 'timeout'
-        self.id = id_
-
-    @classmethod
-    def from_string(cls, node_string):
-        raise NotImplementedError
-
-    def __str__(self):
-        return 'IQ with id %s reached timeout' % self.id
-
-    def serialize(self):
-        raise NotImplementedError
-
-
 class StreamError(CommonError):
     def __init__(self, stanza):
         self.condition = stanza.getError()
@@ -682,7 +649,8 @@ class Properties:
 
 
 class MessageProperties:
-    def __init__(self):
+    def __init__(self, own_jid):
+        self._own_jid = own_jid
         self.carbon = None
         self.type = MessageType.NORMAL
         self.id = None
@@ -729,6 +697,14 @@ class MessageProperties:
         self.security_label = None
         self.chatstate = None
 
+    def is_from_us(self, bare_match=True):
+        if self.from_ is None:
+            raise ValueError('from attribute missing')
+
+        if bare_match:
+            return self._own_jid.bare_match(self.from_)
+        return self._own_jid == self.from_
+
     @property
     def has_user_delay(self):
         return self.user_timestamp is not None
@@ -846,7 +822,8 @@ class MessageProperties:
 
 
 class IqProperties:
-    def __init__(self):
+    def __init__(self, own_jid):
+        self._own_jid = own_jid
         self.type = None
         self.jid = None
         self.id = None
@@ -855,6 +832,7 @@ class IqProperties:
         self.payload = None
         self.http_auth = None
         self.ibb = None
+        self.blocking = None
 
     @property
     def is_http_auth(self):
@@ -864,9 +842,14 @@ class IqProperties:
     def is_ibb(self):
         return self.ibb is not None
 
+    @property
+    def is_blocking(self):
+        return self.blocking is not None
+
 
 class PresenceProperties:
-    def __init__(self):
+    def __init__(self, own_jid):
+        self._own_jid = own_jid
         self.type = None
         self.priority = None
         self.show = None
@@ -931,9 +914,7 @@ class PresenceProperties:
     def new_jid(self):
         if not self.is_nickname_changed:
             raise ValueError('This is not a nickname change')
-        jid = JID(self.jid)
-        jid.setResource(self.muc_user.nick)
-        return jid
+        return self.jid.new_with(resource=self.muc_user.nick)
 
     @property
     def is_kicked(self):
